from .__ import ____
from .__ import *

from ._AI import VAD, Whisper
from ._Audio import Audio
from ._Texture import Texture


class A_Roll:
	_vad: VAD = None
	_whisp: Whisper = None

	_rawFiPa: Path = None
	_proxyFiPa: Path = None
	_draftFiPa: Path = None
	_proofFiPa: Path = None
	_optimFiPa: Path = None

	_dummyFiPa: Path = None

	_proxy: Audio = None
	_draft: Texture = None
	_proof: Texture = None
	_optim: Texture = None

	def __init__(self, taskDiPa: Path, vad: Optional[VAD], whisp: Optional[Whisper]):
		self._vad = vad
		self._whisp = whisp

		task = taskDiPa.name
		assert taskDiPa.is_dir()
		self._rawFiPa = [_ for _ in taskDiPa.iterdir() if _.stem == task + '┆0_Raw'][0]
		self._proxyFiPa = taskDiPa / (task + '┆1_Proxy.aac')
		self._draftFiPa = taskDiPa / (task + '┆2_Draft.srt')
		self._proofFiPa = taskDiPa / (task + '┆3_Proof.srt')
		self._optimFiPa = taskDiPa / (task + '┆4_Optim.srt')

		self._dummyFiPa = taskDiPa / (task + '┆X_Dummy.mp4')
		pass

	def loadAI(self):
		if not self._vad:
			print('[VAD]')
			self._vad = VAD()
			pass
		if not self._whisp:
			print('[Whisper]')
			self._whisp = Whisper('large')
			pass
		pass

	def genProxy(self):
		if not self._proxyFiPa.is_file():
			____.genProxy(self._rawFiPa, self._proxyFiPa)
			pass
		pass

	def loadProxy(self):
		if not self._proxy:
			self.genProxy()
			self._proxy = Audio.load(self._proxyFiPa).normLufs(____.LUFS)
			pass
		pass

	def genDraft(self):
		if not self._draftFiPa.is_file():
			self.loadProxy()
			self.loadAI()

			cont: Texture = self._vad.probe(self._proxy)
			cont = cont.aggrByGran()

			draft = self._whisp.transcribe(self._proxy, cont)
			draft = draft.calibrate(self._proxy, -60)
			draft = draft.extent([100, 0])
			draft.save(self._draftFiPa.as_posix())
			pass
		pass

	def loadDraft(self):
		if not self._draft:
			self.genDraft()
			self._draft = Texture.load(self._draftFiPa.as_posix())
			pass
		pass

	def loadProof(self):
		if self._proofFiPa.is_file():
			self._proof = Texture.load(self._proofFiPa.as_posix())
			pass
		else:
			print('!!! Proof not exists...')
			pass
		pass

	def genOptim(self, accent: str = None):
		if not self._optimFiPa.is_file():
			self.loadProxy()
			self.loadProof()
			optim = self._proof.calibrate(self._proxy, -50)
			optim = optim.extent([100, 50])
			if accent:
				optim = optim.toAccent(accent)
				pass
			optim.save(self._optimFiPa.as_posix())
			pass
		pass

	def loadOptim(self):
		if not self._optim:
			self.genOptim()
			self._optim = Texture.load(self._optimFiPa.as_posix())
			pass
		pass

	def genDummy(self):
		if not self._dummyFiPa.is_file():
			____.genDummy(self._rawFiPa, self._dummyFiPa)
			pass
		pass

	pass



from .__ import *

from ._Stamp import Stamp
from ._Audio import Audio
from ._Texture import Texture


class VAD:
	_model, _tools = torch.hub.load(
		repo_or_dir='snakers4/silero-vad',
		model='silero_vad',
		verbose=False
	)

	def probe(
		self,
		audio: Audio,
		thresh: float = 0.8,
		minGran: int = 150,
		minGap: int = 100,
		pad: int = 200,
	):
		print('<probe>')
		fStamps: list[dict] = self._tools[0](
			audio.sig,
			self._model,
			sampling_rate=audio.fps,
			threshold=thresh,
			min_speech_duration_ms=minGran,
			min_silence_duration_ms=minGap,
			speech_pad_ms=pad,
			window_size_samples=1024,
		)
		___ = Texture([SSAEvent(*Stamp(*_.values(), unit=audio.fps)) for _ in fStamps])
		return ___

	pass


class Whisper:
	def __init__(self, model: str):
		self.model = whisper.load_model(model)
		pass

	def transcribe(
		self,
		audio: Audio,
		cont: Texture,
		lang: str = 'Chinese',
		accent: str = 'zh-cn',
	):
		print('<transcribe>')
		___ = Texture()
		prompt = None

		for contEvent in cont:
			clip = audio.pick(contEvent)
			res = self.model.transcribe(
				clip.sig,
				language=lang,
				prompt=prompt,
				no_speech_threshold=0.2,
				temperature=0,
			)

			prompt = ''
			for seg in res['segments']:
				msStamp = [round(seg[_] * 1000) for _ in ['start', 'end']]
				segEvent = SSAEvent(*msStamp, seg['text'])
				segEvent.shift(ms=contEvent.start)

				t = min(segEvent.end, contEvent.end)
				if segEvent.start < t:
					segEvent.end = t
					segEvent.text = regex.sub('[,?，。？]', ' ', segEvent.text)
					segEvent.text = segEvent.text.strip()
					if accent:
						segEvent.text = zhconv.convert(segEvent.text, accent)
						pass
					___.append(segEvent)
					prompt += segEvent.text
					pass
				pass

			prompt = prompt.replace(' ', '')
			pass
		return ___

	pass



from .__ import ____
from .__ import *

from ._Stamp import Stamp


class Audio:
	sig: numpy.ndarray = None
	fps: int = None

	def __init__(self, sig: numpy.ndarray, fps: int):
		self.sig = sig.copy()
		self.fps = fps
		pass

	@classmethod
	def load(cls, mediaFiPa: Path, fps: int = ____.FPS):
		print('<load>')
		sig, fps = librosa.load(mediaFiPa.as_posix(), sr=fps)
		return cls(sig, fps)

	def __getitem__(self, stamp: Stamp):
		fStamp = [round(_ / 1000 * self.fps) for _ in stamp]
		return Audio(self.sig[slice(*fStamp)], self.fps)

	def pick(self, event: SSAEvent):
		return self[Stamp(event.start, event.end)]

	def dura(self):
		return round(len(self.sig) / self.fps * 1000)

	def save(self, oFiPa: Path) -> None:
		soundfile.write(oFiPa, self.sig, self.fps)
		pass

	def trim(self, db: float):
		_, fStamp = librosa.effects.trim(
			self.sig,
			top_db=-db,
			ref=1,
			frame_length=256,
			hop_length=64,
		)
		audio = Audio(_, self.fps)
		stamp = Stamp(*fStamp, unit=self.fps)
		return audio, stamp

	def pad(self, targDura: int = 20000):
		targLen = round(targDura / 1000 * self.fps)
		padLen = max(0, targLen - len(self.sig))
		sig = numpy.pad(self.sig, (0, padLen))
		return Audio(sig, self.fps)

	def normPeak(self, peak: float = -1):
		print('<normPeak>')
		sig = pyloudnorm.normalize.peak(self.sig, peak)
		return Audio(sig, self.fps)

	def normLufs(self, lufs: float = -15):
		print('<normLufs>')
		curr = pyloudnorm.Meter(self.fps).integrated_loudness(self.sig)
		sig = pyloudnorm.normalize.loudness(self.sig, curr, lufs)
		return Audio(sig, self.fps)

	pass



from .__ import ____
from .__ import *

from ._Audio import Audio
from ._Stamp import Stamp


class Texture(SSAFile):

	def __init__(self, tex: Union[SSAFile, list[SSAEvent]] = None):
		super().__init__()
		if tex:
			self.extend([_.copy() for _ in tex])
		pass

	def aggrByGran(self, maxGran: int = ____.MAX_GRAN):
		if len(self) <= 1:
			___ = Texture(self)
			pass
		elif self[-1].end - self[0].start < maxGran:
			text = '\n'.join([_.text for _ in self])
			event = SSAEvent(self[0].start, self[-1].end, text)
			___ = Texture([event])
			pass
		else:
			key = lambda _: self[_].start - self[_ - 1].end
			theIdx = max(range(1, len(self)), key=key)

			_L = Texture(self[:theIdx])
			_L = _L.aggrByGran(maxGran)

			_R = Texture(self[theIdx:])
			_R = _R.aggrByGran(maxGran)

			___ = Texture(_L.events + _R.events)
			pass
		return ___

	def aggrByGap(self, minGap: int = ____.MIN_GAP):
		___ = Texture()
		lastEnd = -minGap - 1
		for event in self:
			if event.start - lastEnd > minGap:
				___.append(event.copy())
				pass
			else:
				___[-1].end = event.end
				___[-1].text = '\n'.join([___[-1].text, event.text])
				pass
			lastEnd = ___[-1].end
			pass
		return ___

	def calibrate(self, audio: Audio, db: float, step: int = ____.STEP):
		print('<calibrate>')
		DURA = audio.dura()
		___ = Texture(self)
		N = len(___)

		for i in range(N):
			cap = [
				0 if i == 0 else ___[i - 1].end,
				DURA if i == N - 1 else ___[i + 1].start,
			]
			lastDura = 0
			done = False
			while not done:
				stamp_1 = Stamp(
					max(cap[0], ___[i].start - step),
					min(cap[1], ___[i].end + step),
				)
				_, stamp_2 = audio[stamp_1].trim(db)
				stamp_2 = stamp_2.shift(stamp_1[0])

				dura = stamp_2[1] - stamp_2[0]
				done = (dura <= lastDura)
				if not done:
					___[i].start, ___[i].end = stamp_2
					lastDura = dura
					pass
				pass
			pass
		return ___

	def extent(self, ex: list[int]):
		print('<extent>')
		___ = Texture(self)
		N = len(___)

		___[0].start = max(___[0].start - ex[0], 0)
		___[N - 1].end = ___[N - 1].end + ex[1]

		for i in range(1, N):
			___[i].start = max(___[i].start - ex[0], ___[i - 1].end)
			___[i - 1].end = min(___[i - 1].end + ex[1], ___[i].start)
			pass
		return ___

	def toAccent(self, accent: str):
		print('<toAccent>')
		___ = Texture(self)
		for event in ___:
			event.text = zhconv.convert(event.text, accent)
			pass
		return ___

	pass



import json5
import numpy
import torch
import zhconv
import ffmpeg
import whisper
import librosa
import soundfile
import pyloudnorm
import subprocess

import re as regex

from pathlib import Path

from typing import Optional, Union
from pysubs2 import SSAFile, SSAEvent

import xml.etree.ElementTree as ET

import opentimelineio as otio
import opentimelineio.opentime as ot
from opentimelineio.schema import Clip, Timeline, ExternalReference, MissingReference


class ____:
	FPS: int = 16000

	LUFS: float = -15
	TP: float = -0.5

	MAX_GRAN: int = 20000
	MIN_GAP: int = 50

	STEP = 10

	@classmethod
	def genProxy(
		cls,
		rawFiPa: Path, proxyFiPa: Path,
		fps: int = FPS, lufs: float = LUFS, tp: float = TP
	):
		print('<genProxy>')
		pipe = ffmpeg.input(rawFiPa.as_posix())
		pipe = ffmpeg.filter(pipe, "loudnorm", I=lufs, TP=tp)
		pipe = ffmpeg.output(pipe, proxyFiPa.as_posix(), ar=fps, ac=1)
		ffmpeg.run(pipe, quiet=True)
		pass

	@classmethod
	def genDummy(cls, rawFiPa: Path, dummyFiPa: Path):
		print('<genDummy>')
		cmd = ' '.join([
			'ffmpeg -v warning',
			'-f lavfi -i color=c=0x202020:s=1920x1080:r=10',
			'-i "%s"' % rawFiPa.as_posix(),
			'-c:a copy',
			'-shortest',
			'"%s"' % dummyFiPa.as_posix(),
		])
		print(cmd)
		subprocess.run(cmd, encoding='UTF-8')
		pass

	@classmethod
	def getMediaInfo(cls, mediaFiPa: Path):
		pass

	pass



from .__ import *


class Stamp(list[int, int]):

	def __init__(self, t0, t1, unit: Union[str, int] = 'ms'):
		___ = [t0, t1]
		if isinstance(unit, str):
			unit = unit.lower()
			assert unit in {'ms', 's'}
			if unit == 's':
				___ = [_ * 1000 for _ in ___]
				pass
			pass
		else:
			___ = [_ * 1000 / unit for _ in ___]
			pass
		___ = [round(_) for _ in ___]
		super().__init__(___)
		pass

	def shift(self, ms):
		return [_ + ms for _ in self]

	pass
